/**
 * test-webid
 * Copyright (C) 2011 Thomas Bergwinkl
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bergnet.webid.commons;

import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import javax.security.auth.x500.X500Principal;

import org.bouncycastle.asn1.ASN1Encodable;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.x509.X509V3CertificateGenerator;


public class WebID {
	private PrivateKey privateKey;
	private X509Certificate certificate;
	private ArrayList<String> uris;

	@SuppressWarnings("rawtypes")
	public WebID(PrivateKey privateKey, X509Certificate certificate) throws WebIDException {
		this.privateKey = privateKey;
		this.certificate = certificate;
		this.uris = new ArrayList<String>();

		try {
			for(Object sanObject : certificate.getSubjectAlternativeNames()) {
				if(sanObject instanceof List && ((List)sanObject).size() >= 2) {				
					List sanObjectList = (List)sanObject;

					if(sanObjectList.get(0).equals(GeneralName.uniformResourceIdentifier))
						this.uris.add(sanObjectList.get(1).toString());
				}
			}
		} catch(CertificateParsingException e) {
			throw new WebIDException("Couldn't read SubjectAlternativeNames", e);
		}
	}

	public PrivateKey getPrivateKey() {
		return privateKey;
	}

	public X509Certificate getCertificate() {
		return certificate;
	}

	public String[] getUris() {
		return uris.toArray(new String[]{});
	}

	public BigInteger getModulus() {
		if(!(this.certificate.getPublicKey() instanceof RSAPublicKey))
			return null;

		RSAPublicKey rsaPublicKey = (RSAPublicKey)this.certificate.getPublicKey();

		return rsaPublicKey.getModulus();
	}

	public BigInteger getPublicExponent() {
		if(!(this.certificate.getPublicKey() instanceof RSAPublicKey))
			return null;

		RSAPublicKey rsaPublicKey = (RSAPublicKey)this.certificate.getPublicKey();

		return rsaPublicKey.getPublicExponent();
	}
	
	public static KeyPair generateRsaKeyPair(int keySize) throws WebIDException {
		initBouncyCastleProvider();

		KeyPairGenerator keyPairGenerator;

		try {
			keyPairGenerator = KeyPairGenerator.getInstance("RSA", "BC");
		} catch (Exception e) {
			throw new WebIDException("Couldn't access rsa key pair generator", e);
		}

		keyPairGenerator.initialize(keySize, new SecureRandom());

		return keyPairGenerator.generateKeyPair();		
	}

	public static WebID generate(KeyPair keyPair, String[] uris, Date validFrom, Date validTill) throws WebIDException {
		initBouncyCastleProvider();

		X509V3CertificateGenerator certificateGenerator = new X509V3CertificateGenerator();

		certificateGenerator.setSerialNumber(BigInteger.valueOf(validFrom.getTime()));
		certificateGenerator.setIssuerDN(new X500Principal("CN=Me"));
		certificateGenerator.setNotBefore(validFrom);
		certificateGenerator.setNotAfter(validTill);
		certificateGenerator.setSubjectDN(new X500Principal("CN=Me"));
		certificateGenerator.setPublicKey(keyPair.getPublic());
		certificateGenerator.setSignatureAlgorithm("SHA256WithRSAEncryption");

		ArrayList<ASN1Encodable> sanList = new ArrayList<ASN1Encodable>();

		for(String uri : uris)
			sanList.add(new GeneralName(GeneralName.uniformResourceIdentifier, uri));
		
	    try {
	    	/*certificateGenerator.addExtension(
	    			X509Extensions.BasicConstraints,
	    			true,
	    			new BasicConstraints(false));
	    	certificateGenerator.addExtension(
	    			X509Extensions.KeyUsage,
	    			true,
	    			new KeyUsage(KeyUsage.cRLSign | KeyUsage.digitalSignature | KeyUsage.keyAgreement | KeyUsage.keyEncipherment | KeyUsage.nonRepudiation));
	    	certificateGenerator.addExtension(
	    			MiscObjectIdentifiers.netscapeCertType,
	    			true,
	    			new NetscapeCertType(NetscapeCertType.smime | NetscapeCertType.sslClient));
	    	certificateGenerator.addExtension(
	    			X509Extensions.SubjectKeyIdentifier,
	    			true,
	    			new SubjectKeyIdentifierStructure(keyPair.getPublic()));*/
	    	/*certificateGenerator.addExtension(
	    			X509Extensions.SubjectAlternativeName,
	    			true,
	    			new GeneralNames(new GeneralName(GeneralName.uniformResourceIdentifier, uriStringBuffer.toString())));*/

	    	certificateGenerator.addExtension(
	    			X509Extensions.SubjectAlternativeName,
	    			true,
	    			new GeneralNames(new DERSequence(sanList.toArray(new ASN1Encodable[]{}))));

			return new WebID(keyPair.getPrivate(), certificateGenerator.generate(keyPair.getPrivate(), new SecureRandom()));
		} catch (Exception e) {
			throw new WebIDException("Couldn't generate x509 certificate", e);
		}
	}

	public static WebID generate(int keySize, String[] uris, Date validFrom, Date validTill) throws WebIDException {
		KeyPair keyPair = generateRsaKeyPair(keySize);

		return generate(keyPair, uris, validFrom, validTill);
	}

	public static WebID generate(int keySize, String uri, Date validFrom, Date validTill) throws WebIDException {
		return generate(keySize, new String[] { uri }, validFrom, validTill);
	}

	public static Date createDeltaDate(int days) {
		Calendar calendar = Calendar.getInstance();
		calendar.add(Calendar.DAY_OF_YEAR, days);

		return calendar.getTime();
	}

	private static void initBouncyCastleProvider() {
		if(Security.getProvider("BC") == null)
			Security.addProvider(new BouncyCastleProvider());
	}
}
